@building-option +create-set-co
@create $utils named dictionary utilities:dictionary utilities,dict utilities
@prop _."_waif_class" #5203 rc
@prop _."_perversity" 5 rc
@prop _."empty" "replace with empty dict waif" rc
@prop _."help_text" {} rc
;_.("help_text") = {"See \"look $dict\" for a summary of dictionary functions.", "", "A dictionary is a set of mappings between keys and values.  Dictionaries are mostly like lists, except dictionary indexes aren't restricted to numbers.", "", "Sometimes other languages refer to this data structure as \"table\", \"hash\", \"associative array\", \"partial function\", etc.", "", "Here are some code snippets showing how you can use them:", "", "Initialization:", "  this.last_message = $dict.empty;", "", "Adding", "  this.last_message[player] = message;", "", "Presence:", "  if ($dict:contains(this.last_message, player))", "    player:tell(\"Your last message was \", this.last_message[player]);", "  else", "    player:tell(\"You have no last message.\");", "  endif", "", "Removing items:", "  player:tell(\"Removing your message.\");", "  this.last_message = $dict:remove(this.last_message, player);", "", "Iteration:", "  for pair in ($dict:pairs(this.last_message))", "    {key, value} = pair;", "    if ($string_utils:is_uppercase(value))", "      player:tell(key:dnamec(), \" shouted: \", value);", "    endif", "  endfor", "", "Note that dictionaries have no particular order; :keys and :pairs may return a different order than you added their entries.  (In fact, the current implementation perversely shuffles its internal order to prevent inadvertent code dependencies on ordering.)", "", "Like all MOO values (except waifs), dictionaries are immutable.  As with lists, \"d[k] = v\" will only modify the variable \"d\", and no other dictionaries.", "", "On Waterpoint, dictionaries are currently implemented as waifs, with special server support for directing \"d[k]\" and \"d[k] = v\" to verbs on the waif class.  If performance suggests it, the $dict interface could transparently use multiple waif classes for different sizes/usage patterns. It should be possible to build an implementation of the $dict interface using lists (for MOOs without waifs), although those servers will have to use $dict:get() and $dict:set() to access them.  Eventually, the MOO server may get a built-in dictionary type, and this $dict interface should remain upward compatible."}
"_.("key") => E_PERM (Permission denied)
;_.("aliases") = {"dictionary utilities", "dict utilities"}
;_.("description") = {"See \"help $dict\" for background information on dictionaries.", "", "Accessing", "=========", "", "$dict:is_dict(x) => true iff x is a dictionary", "$dict:contains(dict, key) => true iff dict has key present", "$dict:get(dict, key) => value or raises E_RANGE if key not present", "dict[key] => value or raises E_RANGE", "$dict:get_default(dict, key, default) => value, or default if key not present", "", "Modifying", "=========", "", "$dict:add_or_set(dict, key, value) => dictionary, with key mapped to value.", "  (also callable as $dict:set)", "d[key] = value => d becomes new dictionary, with key mapped to value", "", "$dict:strict_add(dict, key, value) => dictionary, with key mapped to value.", "  if key is already present, raises E_RANGE", "", "$dict:strict_set(dict, key, value) => dictionary, with key mapped to value.", "  if key is not present in dict, raises E_RANGE", "", "", "Working with pairs", "==================", "", "$dict:keys(dict) => list of keys (in arbitrary order)", "$dict:pairs(dict) => {{key, value}, {key, value}, ...} (in arbitrary order)", "$dict:plist(dict) => {{key, key, ...}, {value, value, ...}}(in arbitrary order)", "", "Creating", "========", "", "$dict.empty => The empty dictionary", "$dict:new() => The empty dictionary", "$dict:from_pairs({{key, value}, {key, value}, ...}) => dictionary", "$dict:from_plist({{key, key, ...}, {value, value, ...}}) => dictionary"}

@verb _:"contains get get_default remove keys pairs plist add_or_set strict_add strict_set set values" this none this
@program _:contains
{dict, @extra} = args;
this:_check_type(dict);
return dict:(verb)(@extra);
.

@verb _:"_check_type" this none this
@program _:_check_type
dict = args[1];
if ((typeof(dict) != $waif.type) || (dict.class != this._waif_class))
  raise(E_TYPE);
endif
.

@verb _:"is_dict" this none this rxd
@program _:is_dict
"Usage:  :is_dict(v)";
"";
v = args[1];
return (typeof(v) == $waif.type) && (v.class == this._waif_class);
.

@verb _:"from_plist" this none this rxd
@program _:from_plist
"Usage:  :from_plist(keys, values)";
"";
keys = args[1];
values = args[2];
return this._waif_class:new(keys, values);
.

@verb _:"from_pairs" this none this rxd
@program _:from_pairs
"Usage:  :from_alist(alist)";
"";
"Paranoid code checks this is really a list of pairs instead of using $list_utils:slice.";
alist = args[1];
keys = {};
values = {};
for pair in (alist)
  {key, value} = pair;
  if (random(this._perversity) == 1)
    keys = {key, @keys};
    values = {value, @values};
  else
    keys = {@keys, key};
    values = {@values, value};
  endif
endfor
return this._waif_class:new(keys, values);
.

@verb _:"_implement_perversity" this none this rxd
@program _:_implement_perversity
"Usage:  :_implement_perversity(keys, values)";
"";
keys = args[1];
values = args[2];
keylen = length(keys);
e1 = random(keylen);
e2 = random(keylen);
if (e1 != e2)
  tk = keys[e1];
  tv = values[e1];
  keys[e1] = keys[e2];
  values[e1] = values[e2];
  keys[e2] = tk;
  values[e2] = tv;
endif
return {keys, values};
.

@verb _:"new empty" this none this rxd
@program _:new
return this._waif_class:new();
.

"***finished***
